from more_itertools import powerset


def powerset_max_length(candidates, length):
    return filter(lambda s: len(s) <= length, powerset(candidates))


def no_duplicates_generator_decorator(func):
    def inner(*args, **kwargs):
        seen = set()
        for item in func(*args, **kwargs):
            if item not in seen:
                yield item
                seen.add(item)

    return inner



import logging
from itertools import combinations
from itertools import product

from mlscraper.matches import DictMatch
from mlscraper.matches import ListMatch
from mlscraper.matches import ValueMatch
from mlscraper.samples import TrainingSet
from mlscraper.scrapers import DictScraper
from mlscraper.scrapers import ListScraper
from mlscraper.scrapers import ValueScraper
from mlscraper.selectors import generate_unique_selectors_for_nodes
from mlscraper.selectors import PassThroughSelector
from more_itertools import first
from more_itertools import flatten
from more_itertools import unzip


class TrainingException(Exception):
    pass


class NoScraperFoundException(TrainingException):
    pass


def get_match_combination_priority(matches):
    # check for similarity between matches
    return sum(m1.get_similarity_to(m2) for m1, m2 in combinations(matches, 2))


def train_scraper(training_set: TrainingSet, complexity=100):
    """
    Train a scraper able to extract the given training data.
    """

    logging.info(f"training {training_set=}")

    logging.info(
        "number of matches found per sample: %s",
        [(s, len(s.get_matches())) for s in training_set.item.samples],
    )

    sample_matches = [
        sorted(s.get_matches(), key=lambda m: m.span)[:100]
        for s in training_set.item.samples
    ]
    match_combinations = list(product(*sample_matches))
    logging.info(f"Trying {len(match_combinations)=}")

    # to train quicker, we'll start with combinations that have a high depth
    # this prefers matches, that have a deep root
    # and are thus closer to each other
    match_combinations_prioritized = sorted(
        match_combinations, key=get_match_combination_priority, reverse=True
    )

    for match_combination in match_combinations_prioritized:
        progress_ratio = match_combinations_prioritized.index(match_combination) / len(
            match_combinations_prioritized
        )
        logging.info(f"progress {progress_ratio}")
        try:
            logging.info(f"trying to train scraper for matches ({match_combination=})")
            roots = [s.page for s in training_set.item.samples]
            scraper = train_scraper_for_matches(match_combination, roots, complexity)
            return scraper
        except NoScraperFoundException:
            logging.info(
                "no scraper found "
                "for complexity and match_combination "
                f"({complexity=}, {match_combination=})"
            )
    raise NoScraperFoundException("did not find scraper")


def train_scraper_for_matches(matches, roots, complexity: int):
    """
    Train a scraper that finds the given matches from the given roots.
    :param matches: the matches to scrape
    :param roots: the root elements containing the matches, e.g. pages or elements on pages
    :param complexity: the complexity to try
    """
    found_types = set(map(type, matches))
    assert (
        len(found_types) == 1
    ), f"different match types passed {found_types=}, {matches=}"
    found_type = first(found_types)

    # make sure we have lists
    matches = list(matches)
    roots = list(roots)

    assert len(matches) == len(roots), f"got uneven inputs ({matches=}, {roots=})"

    if len({m.root.soup.name for m in matches}) != 1:
        raise NoScraperFoundException("different names found")

    if any(c1.has_overlap(c2) for c1, c2 in combinations(matches, 2)):
        raise NoScraperFoundException("a pair of matches overlaps, most likely invalid")

    if found_type == ValueMatch:
        logging.info("training ValueScraper")
        matches: list[ValueMatch]

        # if matches have different extractors, we can't find a common scraper
        extractors = set(map(lambda m: m.extractor, matches))
        if len(extractors) != 1:
            raise NoScraperFoundException(
                "different extractors found for matches, aborting"
            )
        extractor = first(extractors)

        # early return: nodes are matched already, e.g. for List of Values
        if all(m.node == r for m, r in zip(matches, roots)):
            # nodes are matched already, done
            return ValueScraper(PassThroughSelector(), extractor=extractor)
        else:
            logging.info(
                "no early return: %s",
                [(m.node, r, m.node == r) for m, r in zip(matches, roots)],
            )

        selector = first(
            generate_unique_selectors_for_nodes(
                [m.node for m in matches], roots, complexity
            ),
            None,
        )
        if not selector:
            raise NoScraperFoundException(f"no selector found {matches=}")
        logging.info(f"found selector for ValueScraper ({selector=})")
        return ValueScraper(selector, extractor)
    elif found_type == DictMatch:
        logging.info("training DictScraper")
        matches: list[DictMatch]

        # what if some matches have missing keys? idk
        # by using union of all keys, we'll get errors two lines below to be sure
        keys = set(flatten(m.match_by_key.keys() for m in matches))

        # train scraper for each key of dict
        # matches are the matches for the keys
        # roots are the original roots(?)
        scraper_per_key = {}
        for k in keys:
            matches_per_key = [m.match_by_key[k] for m in matches]
            logging.info(f"training key for DictScraper ({k=})")
            logging.info(f"matches for key: {matches_per_key=}")
            scraper = train_scraper_for_matches(matches_per_key, roots, complexity)
            scraper_per_key[k] = scraper
        logging.info(f"found DictScraper ({scraper_per_key=})")
        return DictScraper(scraper_per_key)
    elif found_type == ListMatch:
        logging.info("training ListScraper")
        matches: list[ListMatch]
        logging.info(matches)

        # so we have a list of ListMatch objects
        # we have to find a selector that uniquely matches the list elements
        # todo can be one of the parents
        # for each match, generate all the nodes of list items
        list_item_match_and_roots = [
            (im, r) for m, r in zip(matches, roots) for im in m.matches
        ]
        list_item_nodes_and_roots = [
            (im.root, r) for im, r in list_item_match_and_roots
        ]
        item_nodes, item_roots = unzip(list_item_nodes_and_roots)

        # first selector is fine as it matches perfectly
        # no need to try other selectors
        # -> item_scraper would be the same
        selector = first(
            generate_unique_selectors_for_nodes(
                list(item_nodes), list(item_roots), complexity
            ),
            None,
        )
        if selector:
            logging.info(f"selector that matches list items found ({selector=})")
            # so we have found a selector that matches the list items
            # we now need a scraper, that scrapes each contained item
            # todo im.root does not hold for all items, could be a parent
            item_matches_and_item_roots = [
                (im, im.root) for im, r in list_item_match_and_roots
            ]
            logging.info(
                f"training to extract list items now ({item_matches_and_item_roots})"
            )
            item_matches, item_roots = unzip(item_matches_and_item_roots)
            item_scraper = train_scraper_for_matches(
                list(item_matches), list(item_roots), complexity
            )
            return ListScraper(selector, item_scraper)
        else:
            raise NoScraperFoundException()
    else:
        raise RuntimeError(f"type not matched: {found_type}")



import logging
import typing
from itertools import product

from mlscraper.html import Page
from mlscraper.matches import DictMatch
from mlscraper.matches import generate_all_value_matches
from mlscraper.matches import is_dimensions_match
from mlscraper.matches import is_disjoint_match_combination
from mlscraper.matches import ListMatch


class ItemStructureException(Exception):
    pass


class NoMatchFoundException(Exception):
    pass


class Sample:
    def __init__(self, page: Page, value: typing.Union[str, list, dict]):
        self.page = page
        self.value = value

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.page=}, {self.value=}>"

    def get_matches(self):
        # todo: fix creating new sample objects, maybe by using Item class?

        if isinstance(self.value, str):
            # generate all matches
            value_matches = list(generate_all_value_matches(self.page, self.value))

            # filter out dimensions like width/height
            value_matches = [vm for vm in value_matches if not is_dimensions_match(vm)]

            # raise if not found
            logging.info(
                f"found {len(value_matches)=} on page ({self.value=}, {self.page=})"
            )
            logging.info(f"{value_matches=}")
            if not value_matches:
                raise NoMatchFoundException(
                    f"No match found on page ({self.page=}, {self.value=})"
                )
            return value_matches

        if isinstance(self.value, list):
            matches_by_value = [Sample(self.page, v).get_matches() for v in self.value]

            # generate list of combinations
            # todo filter combinations that use the same matches twice
            # todo create combinations only in order
            match_combis = product(*matches_by_value)

            return [
                ListMatch(tuple(match_combi))
                for match_combi in match_combis
                if is_disjoint_match_combination(match_combi)
            ]

        if isinstance(self.value, dict):
            matches_by_key = {
                k: Sample(self.page, self.value[k]).get_matches() for k in self.value
            }

            return [
                DictMatch(dict(zip(matches_by_key.keys(), mc)))
                for mc in product(*matches_by_key.values())
                if is_disjoint_match_combination(mc)
            ]

        raise RuntimeError(f"unsupported value: {self.value}")


class TrainingSet:
    """
    This class turn samples into an item structure to scrape later.
    """

    item = None

    def add_sample(self, sample: Sample):
        if not self.item:
            self.item = Item.create_from(sample.value)

        self.item.add_sample(sample)

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.item=}>"


class Item:
    """
    An item represents the data structure to scrape, e.g. a specific dict with keys a, b, and c.
    """

    samples = None

    @classmethod
    def create_from(cls, item):
        if isinstance(item, str):
            return ValueItem()
        elif isinstance(item, list):
            return ListItem()
        elif isinstance(item, dict):
            return DictItem()
        else:
            raise ItemStructureException(
                f"unsupported item type ({item=}, {type(item)=}"
            )

    def __init__(self):
        self.samples = []

    def add_sample(self, sample: Sample):
        self.samples.append(sample)

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.samples=}>"


class DictItem(Item):
    item_per_key = None

    def __init__(self):
        super().__init__()
        self.item_per_key = {}

    def add_sample(self, sample: Sample):
        if not isinstance(sample.value, dict):
            raise ItemStructureException(f"dict expected, {sample.value} given")

        super().add_sample(sample)

        for key, value in sample.value.items():
            if key not in self.item_per_key:
                self.item_per_key[key] = Item.create_from(value)

            value_sample = Sample(sample.page, value)
            self.item_per_key[key].add_sample(value_sample)


class ListItem(Item):
    item = None

    def __init__(self):
        super().__init__()
        self.item = None

    def add_sample(self, sample: Sample):
        if not isinstance(sample.value, list):
            raise ItemStructureException(f"list expected, {sample.value} given")

        super().add_sample(sample)

        if not self.item and len(sample.value):
            self.item = Item.create_from(sample.value[0])

        for v in sample.value:
            self.item.add_sample(Sample(sample.page, v))


class ValueItem(Item):
    def add_sample(self, sample: Sample):
        if not isinstance(sample.value, str):
            raise ItemStructureException(f"str expected, {sample.value} given")
        super().add_sample(sample)


def make_training_set(pages, items):
    assert len(pages) == len(items)

    ts = TrainingSet()
    for p, i in zip(pages, items):
        ts.add_sample(Sample(p, i))

    return ts



from mlscraper.html import Node
from mlscraper.matches import Extractor
from mlscraper.selectors import Selector


class Scraper:
    def get(self, node: Node):
        raise NotImplementedError()


class DictScraper(Scraper):
    scraper_per_key = None

    def __init__(self, scraper_per_key: dict[str, Scraper]):
        self.scraper_per_key = scraper_per_key

    def get(self, node: Node):
        return {key: scraper.get(node) for key, scraper in self.scraper_per_key.items()}

    def __repr__(self):
        return f"<DictScraper {self.scraper_per_key=}>"


class ListScraper(Scraper):
    selector = None
    scraper = None

    def __init__(self, selector: Selector, scraper: Scraper):
        self.selector = selector
        self.scraper = scraper

    def get(self, node: Node):
        return [
            self.scraper.get(item_node) for item_node in self.selector.select_all(node)
        ]

    def __repr__(self):
        return f"<ListScraper {self.selector=} {self.scraper=}>"


class ValueScraper(Scraper):
    selector = None
    extractor = None

    def __init__(self, selector: Selector, extractor: Extractor):
        self.selector = selector
        self.extractor = extractor

    def get(self, node: Node):
        return self.extractor.extract(self.selector.select_one(node))

    def __repr__(self):
        return f"<ValueScraper {self.selector=}, {self.extractor=}>"



import logging
from itertools import product

from mlscraper.html import Node
from mlscraper.util import no_duplicates_generator_decorator
from more_itertools import first
from more_itertools import powerset


class Selector:
    """
    Class to select nodes from another node.
    """

    def select_one(self, node: Node) -> Node:
        raise NotImplementedError()

    def select_all(self, node: Node) -> list[Node]:
        raise NotImplementedError()


class PassThroughSelector(Selector):
    def select_one(self, node: Node) -> Node:
        return node

    def select_all(self, node: Node) -> list[Node]:
        # this does not make sense as we have only one node to pass through
        raise RuntimeError("cannot apply select_all to PassThroughSelector")


class CssRuleSelector(Selector):
    def __init__(self, css_rule):
        self.css_rule = css_rule

    def select_one(self, node: Node):
        selection = node.select(self.css_rule)
        if not selection:
            raise AssertionError(
                f"css rule does not match any node ({self.css_rule=}, {node=})"
            )
        return selection[0]

    def select_all(self, node: Node):
        return node.select(self.css_rule)

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.css_rule=}>"


def generate_unique_selectors_for_nodes(nodes: list[Node], roots, complexity: int):
    """
    generate a unique selector which only matches the given nodes.
    """
    if roots is None:
        logging.info("roots is None, setting roots manually")
        roots = [n.root for n in nodes]

    nodes_per_root = {r: [n for n in nodes if n.has_parent(r)] for r in set(roots)}
    for selector in generate_selectors_for_nodes(nodes, roots, complexity):
        if all(
            selector.select_all(r) == nodes_of_root
            for r, nodes_of_root in nodes_per_root.items()
        ):
            yield selector


@no_duplicates_generator_decorator
def generate_selectors_for_nodes(nodes: list[Node], roots, complexity: int):
    """
    Generate a selector which matches the given nodes.
    """

    logging.info(
        f"trying to find selector for nodes ({nodes=}, {roots=}, {complexity=})"
    )
    assert nodes, "no nodes given"
    assert roots, "no roots given"
    assert len(nodes) == len(roots)

    direct_css_selectors = list(_generate_direct_css_selectors_for_nodes(nodes))
    for direct_css_selector in direct_css_selectors:
        yield CssRuleSelector(direct_css_selector)

    ancestors_below_roots = [
        [p for p in n.parents if p.has_parent(r) and p.tag_name != "html"]
        for n, r in zip(nodes, roots)
    ]
    for ancestors in product(*ancestors_below_roots):
        for ancestor_selector_raw in _generate_direct_css_selectors_for_nodes(
            ancestors
        ):
            # generate refinement selectors for parents
            # e.g. if selectivity of child selector is not enough
            for css_selector_raw in direct_css_selectors:
                css_selector_combined = ancestor_selector_raw + " " + css_selector_raw
                yield CssRuleSelector(css_selector_combined)

                # make parent selector
                if all(node.parent == parent for node, parent in zip(nodes, ancestors)):
                    yield CssRuleSelector(
                        f"{ancestor_selector_raw} > {css_selector_raw}"
                    )


def _generate_direct_css_selectors_for_nodes(nodes: list[Node]):
    # pseudo classes apply to already generated selectors
    # and can thus be applied in retrospect

    # see: https://developer.mozilla.org/en-US/docs/Web/CSS/:nth-child
    for css_selector in _generate_direct_css_selectors_for_nodes_without_pseudo(nodes):
        yield css_selector

    # pull to the end as far as possible
    for css_selector in _generate_direct_css_selectors_for_nodes_without_pseudo(nodes):
        if all(n.tag_name not in ["html", "body"] for n in nodes):
            child_indexes = [n.parent.select(css_selector).index(n) for n in nodes]
            if len(set(child_indexes)) == 1:
                # nth is indexed with 1
                nth = 1 + child_indexes[0]
                yield f"{css_selector}:nth-child({nth})"


def _generate_direct_css_selectors_for_nodes_without_pseudo(nodes: list[Node]):
    # check for same tag name
    is_same_tag = len({n.tag_name for n in nodes}) == 1
    common_tag_name = nodes[0].tag_name if is_same_tag else None
    if common_tag_name:
        yield common_tag_name

    # check for common id
    common_ids = {n.id for n in nodes}
    is_same_id = len(common_ids) == 1
    if is_same_id and None not in common_ids:
        yield "#" + first(common_ids)

    # check for common classes
    common_classes = set.intersection(*[set(n.classes) for n in nodes])
    for class_combination in powerset(common_classes):
        if class_combination:
            css_selector = "".join(map(lambda cl: "." + cl, class_combination))
            yield css_selector

            # if same tag name, also yield tag_name + selector
            if common_tag_name:
                yield common_tag_name + css_selector
        else:
            # empty combination -> ignore
            pass

    # check for common attributes
    # see: https://developer.mozilla.org/en-US/docs/Web/CSS/Attribute_selectors
    if common_tag_name:
        common_attributes = set.intersection(
            *[set(n.html_attributes.keys()) for n in nodes]
        )
        common_attributes_filtered = [
            ca for ca in common_attributes if ca not in ["id", "class", "rel"]
        ]
        for common_attribute in common_attributes_filtered:
            yield f"{common_tag_name}[{common_attribute}]"

            # check for common attribute values
            attribute_values = {n.html_attributes[common_attribute] for n in nodes}
            if len(attribute_values) == 1:
                common_attribute_value = first(attribute_values)
                yield f'{common_tag_name}[{common_attribute}="{common_attribute_value}"]'



"""
Encapsulation of html-related functionality.
BeautifulSoup should only get used here.
"""
import html
import logging
import re
from abc import ABC
from dataclasses import dataclass
from functools import cached_property

from bs4 import BeautifulSoup
from bs4 import NavigableString


@dataclass
class HTMLMatch(ABC):
    node: "Node" = None


@dataclass
class HTMLExactTextMatch(HTMLMatch):
    pass


@dataclass
class HTMLAttributeMatch(HTMLMatch):
    attr: str = None


class Node:
    soup = None
    _page = None

    def __init__(self, soup, page: "Page"):
        self.soup = soup
        self._page = page
        self._hash = None

    @property
    def root(self):
        return self._page

    @cached_property
    def depth(self):
        return self.parent.depth

    @cached_property
    def parent(self):
        return self._page._get_node_for_soup(self.soup.parent)

    @cached_property
    def text(self):
        return self.soup.text

    def find_all(self, item) -> list[HTMLMatch]:
        return list(self._generate_find_all(item))

    def _generate_find_all(self, item):
        assert isinstance(item, str), "can only search for str at the moment"

        # text
        # - since text matches including whitespace, a regex is used
        target_regex = re.compile(r"^\s*%s\s*$" % html.escape(item))
        for soup_node in self.soup.find_all(string=target_regex):
            # use parent node as found text is NaviableString and not Tag
            node = self._page._get_node_for_soup(soup_node.parent)
            yield HTMLExactTextMatch(node)

            for p in node.parents:
                if p.text.strip() == node.text.strip():
                    yield HTMLExactTextMatch(p)

        # attributes
        for soup_node in self.soup.find_all():
            for attr in soup_node.attrs:
                if soup_node[attr] == item:
                    node = self._page._get_node_for_soup(soup_node)
                    yield HTMLAttributeMatch(node, attr)

        # todo implement other find methods

    def has_parent(self, node: "Node"):
        for p in self.soup.parents:
            if p == node.soup:
                return True
        return False

    @cached_property
    def parents(self):
        return [self._page._get_node_for_soup(p) for p in self.soup.parents]

    @property
    def classes(self):
        return self.soup.attrs.get("class", [])

    @property
    def id(self):
        return self.soup.attrs.get("id", None)

    @property
    def tag_name(self):
        return self.soup.name

    @property
    def html_attributes(self):
        return self.soup.attrs

    def select(self, css_selector):
        return [
            self._page._get_node_for_soup(n) for n in self.soup.select(css_selector)
        ]

    def __repr__(self):
        if isinstance(self.soup, NavigableString):
            return f"<{self.__class__.__name__} {self.soup.strip()[:10]=}>"
        return (
            f"<{self.__class__.__name__} {self.soup.name=}"
            f" classes={self.soup.get('class', None)},"
            f" text={''.join(self.soup.stripped_strings)[:10]}...>"
        )

    def __hash__(self):
        if not self._hash:
            self._hash = self.soup.__hash__()
        return self._hash
        # return self.soup.__hash__()
        # return super().__hash__()

    def __eq__(self, other):
        return isinstance(other, Node) and self.soup == other.soup


class Page(Node):
    """
    One page, i.e. one HTML document.
    """

    _node_registry = None

    def __init__(self, html):
        self.html = html
        soup = BeautifulSoup(self.html, "lxml")

        # register node for each soup
        self._node_registry = {soup: self}

        super().__init__(soup, self)

    @property
    def depth(self):
        return 0

    def _get_node_for_soup(self, soup) -> Node:
        if soup not in self._node_registry:
            self._node_registry[soup] = Node(soup, self)
        return self._node_registry[soup]


def get_root_node(nodes: list[Node]) -> Node:
    pages = [n._page for n in nodes]
    assert len(set(pages)) == 1, "different pages found, cannot get a root"

    # generate parent paths from top to bottom
    # [elem, parent, ancestor, root]
    parent_paths = [reversed([n] + n.parents) for n in nodes]

    # start looping from bottom to top
    # zip automatically uses common length
    # -> last element is the first one, where len(nodes) roots to compare exist
    for layer_nodes in reversed(list(zip(*parent_paths))):
        if len(set(layer_nodes)) == 1:
            return layer_nodes[0]
    raise RuntimeError("no root found")


def get_relative_depth(node: Node, root: Node):
    node_parents = list(node.soup.parents)

    # depth of root
    i = node_parents.index(root.soup)

    # depth of element
    j = len(node_parents)

    return j - i


def selector_matches_nodes(root: Node, selector: str, expected: list[Node]):
    """
    Check whether the given selector matches the expected nodes.
    """
    logging.info(
        f"checking if selector matches nodes ({root=}, {selector=}, {expected=})"
    )
    # we care for equality here
    # as selector should match the expected nodes in the exact given order
    # we do this here, as wrapping Nodes can have side effects regarding equality
    return root.soup.select(selector) == [n.soup for n in expected]



"""
Matches are specific elements found on a page that match a sample.
"""
import logging
import typing
from functools import cached_property
from itertools import combinations
from itertools import product

from mlscraper.html import get_relative_depth
from mlscraper.html import get_root_node
from mlscraper.html import HTMLAttributeMatch
from mlscraper.html import HTMLExactTextMatch
from mlscraper.html import Node


class Match:
    """
    Occurrence of a specific sample on a page
    """

    @property
    def root(self) -> Node:
        """
        The lowest element that contains matched elements.
        """
        raise NotImplementedError()

    def has_overlap(self, other_match: "Match"):
        assert isinstance(other_match, Match)

        # early return if different document
        if self.root.root != other_match.root.root:
            return False

        return (
            # overlap if same root node
            self.root == other_match.root
            # or if one is a parent of the other one
            or self.root.has_parent(other_match.root)
            or other_match.root.has_parent(self.root)
        )

    @property
    def depth(self):
        """
        How deep inside the DOM the match is.
        """
        # depth of root compared to document
        return self.root.depth

    @property
    def span(self):
        """
        Heuristic for how big of a subtree the match spans.
        """
        raise NotImplementedError()

    def get_similarity_to(self, match: "Match"):
        raise NotImplementedError()


class Extractor:
    """
    Class that extracts values from a node.
    """

    def extract(self, node: Node):
        raise NotImplementedError()


class TextValueExtractor(Extractor):
    """
    Class to extract text from a node.
    """

    def extract(self, node: Node):
        return node.soup.text.strip()

    def __repr__(self):
        return f"<{self.__class__.__name__}>"

    def __hash__(self):
        # todo each instance equals each other instance,
        #  so this holds,
        #  but it isn't pretty
        return 0

    def __eq__(self, other):
        return isinstance(other, TextValueExtractor)


class AttributeValueExtractor(Extractor):
    """
    Extracts a value from the attribute in an html tag.
    """

    attr = None

    def __init__(self, attr):
        self.attr = attr

    def extract(self, node: Node):
        if self.attr in node.soup.attrs:
            return node.soup[self.attr]

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.attr=}>"

    def __hash__(self):
        return self.attr.__hash__()

    def __eq__(self, other):
        return isinstance(other, AttributeValueExtractor) and self.attr == other.attr


class DictMatch(Match):
    match_by_key = None

    def __init__(self, match_by_key: dict):
        self.match_by_key = match_by_key

    @cached_property
    def root(self) -> Node:
        match_roots = [m.root for m in self.match_by_key.values()]
        return get_root_node(match_roots)

    @cached_property
    def span(self):
        # add span from this root to match root
        return sum(
            m.span + get_relative_depth(m.root, self.root)
            for m in self.match_by_key.values()
        )

    def get_similarity_to(self, match: "Match"):
        assert isinstance(match, self.__class__)
        keys = set(self.match_by_key.keys()).intersection(
            set(match.match_by_key.keys())
        )
        return sum(
            self.match_by_key[key].get_similarity_to(match.match_by_key[key])
            for key in keys
        )

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.match_by_key=}>"


class ListMatch(Match):
    matches = None

    def __init__(self, matches: tuple):
        self.matches = matches

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.matches=}>"

    @cached_property
    def root(self) -> Node:
        return get_root_node([m.root for m in self.matches])

    @cached_property
    def span(self):
        return sum(get_relative_depth(m.root, self.root) + m.span for m in self.matches)

    def get_similarity_to(self, match: "Match"):
        assert isinstance(match, self.__class__)
        return sum(
            lm1.get_similarity_to(lm2)
            for lm1, lm2 in product(self.matches, match.matches)
        )


class ValueMatch(Match):
    node = None
    extractor = None

    def __init__(self, node: Node, extractor: Extractor):
        self.node = node
        self.extractor = extractor

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.node=}, {self.extractor=}>"

    @property
    def root(self) -> Node:
        return self.node

    @property
    def span(self):
        return 0

    def get_similarity_to(self, match: "Match"):
        assert isinstance(match, self.__class__)
        if self.extractor != match.extractor:
            return 0

        if self.node.tag_name != match.node.tag_name:
            return 0

        return 1


def generate_all_value_matches(
    node: Node, item: str
) -> typing.Generator[Match, None, None]:
    logging.info(f"generating all value matches ({node=}, {item=})")
    for html_match in node.find_all(item):
        matched_node = html_match.node
        if isinstance(html_match, HTMLExactTextMatch):
            extractor = TextValueExtractor()
            yield ValueMatch(matched_node, extractor)
        elif isinstance(html_match, HTMLAttributeMatch):
            extractor = AttributeValueExtractor(html_match.attr)
            yield ValueMatch(matched_node, extractor)
        else:
            logging.warning(
                "Cannot deal with HTMLMatch type, ignoring "
                f"({html_match=}, {type(html_match)=}))"
            )


def is_disjoint_match_combination(matches):
    """
    Check if the given matches have no overlap.
    """
    return all(not m1.has_overlap(m2) for m1, m2 in combinations(matches, 2))


def is_dimensions_match(m: Match):
    if not isinstance(m, ValueMatch):
        return False

    if not isinstance(m.extractor, AttributeValueExtractor):
        return False

    return m.extractor.attr in ["width", "height"]



#!/usr/bin/env python
"""The setup script."""
from setuptools import setup


setup(
    name="mlscraper",
    install_requires=[
        "beautifulsoup4",
        "lxml",
        "more-itertools>=8",
    ],
)
